"""Fiatio reads and writes the simple
FIAT file format (dwittman@bell-labs.com,
http://dls.bell-labs.com/fiat/fiat.html).  FIAT is
nice because all the header information looks like a comment
to most programs, so they will treat a FIAT file as
simple multi-column ASCII.

This describes fiat 1.1 format, which is nearly 100%
upwards compatible with fiat 1.0 format.
It is defined as follows:

Lines are separated by newlines.

All values are encoded by replacing newline and other difficult
characters by a percent character (%) followed by a hex code
on writing, and the reverse on reading.

At the top of the file, you have a line identifying the format: "# fiat 1.1"
(regexp: "# fiat 1\.[0-9.]+").

Then, you have any number of header lines beginning with "#".
Header lines are in the form "# attribute = value" (where
white space is optional and can be a mixture of spaces and tabs).
The attribute must match [a-zA-Z_][a-zA-Z_0-9]* .
The value is whatever follows the equals sign, after leading and following
white space is stripped.  If the value begins and ends with
the same quote character ['"], the quotes are also stripped off.
Values may contain any character except newline and the chosen quote.
Note that you must quote or encode a value if it begins or ends with whitespace.

Any other header lines are just treated as comments and ignored.

There may be lines in the header of the form
"# TTYPEn = name" or "#ttype4 = name"
which name the columns of the data (the leftmost column is TTYPE1).
If you don't name the nth column, its name will be "n".
When writing, this module adds an attribute
COL_SEPARATOR which contains the numeric code(s)
(ASCII) of the column separator character(s).
The module also adds a COL_EMPTY attribute with the string used to mark an
blank item.  However, all attributes and names are optional.

Finally, the header is followed by multicolumn ASCII data.
Columns are separated (by default) with any white space,
but if there is a COL_SEPARATOR attribute, it is used instead.
Empty entries for columns should be indicated by whatever code is specified
in COL_EMPTY, if that is set.
Otherwise, if COL_SEPARATOR is set, COL_SEPARATOR strings separate items,
some of which may simply be empty.
(In all cases, a completely blank line is treated as a datum which as all
columns empty.)

If there is no DATE attribute, the write routine adds one, using the current date
and time, in the form ccyy-mm-ddThh:mm:ss (as defined in the NASA FITS format).

This is not quite DWittman FIAT (1.0), which forces value to either be quoted
or to contain no white space. Dwittman FIAT will take a line
in the form "#a=b c", and interpret c as a comment, whereas
fiat 1.1 will interpret "b c" as the value.
However, almost all files will be interpreted the same way.

Here's an example:
# fiat 1.1
# TTYPE1 = b
# TTYPE2 = a
# SAMPRATE = 2.3
# DATE = 2001-09-21T21:32:32
# COL_EMPTY = "%na"
# COL_SEPARATOR = "9"
# Comment1
# Comment2
# b	a
2	1
3	2
3	%na
%na	3
%na	%na
0	1
"""


import gpk_writer
import re
import g_encode
import die
import types

__version__ = "$Revision: 1.53 $"


TABLEN = 8




def _alph(s):
	n = min(len(s), 8)
	o = 0.0
	f = 1.0
	# Not OK for unicode. Sigh.
	for i in range(n):
		f = f/256.0
		o += f * ord(s[i])
	return -o


# def col_order(a, b):
	# lc = cmp(len(str(a[0]))+a[1], len(str(b[0]))+b[1])
	# if lc != 0:
		# return lc
	# return cmp(str(a[0]), str(b[0]))

def col_order(a, b):
	sa = str(a)
	sb = str(b)
	lc = cmp(len(sa), len(sb))
	if lc != 0:
		return lc
	return cmp(sa, sb)


_autogen = re.compile("COL_SEPARATOR$", re.IGNORECASE)
_drop = re.compile("(__NAME_TO_COL|__COLUMNS)$")


def write_array(fd, adata, columns=None, comments=None, hdr=None, sep='\t', numeric=False):
	"""Write a rectangular array in FIAT format.
	Adata is a 2-D numpy array or a sequence of sequences.
	"""
	w = writer(fd, sep=sep)

	if columns is not None:
		w.add_cols(columns)
	if hdr is not None:
		w.headers(hdr)
	if comments is not None:
		for c in comments:
			w.comment(c)
	for i in range(len(adata)):
		w.datavec( adata[i], numeric=numeric )


_autogen = re.compile("TTYPE[0-9]+|COL_EMPTY|COL_SEPARATOR", re.IGNORECASE)

def write(fd, ldata, comments=None, hdr=None, sep='\t', blank='%na', fixed_order=0):
	"""Write a file in FIAT format.
			Ldata is a list of dictionaries.  Each dictionary
			corresponds to one line in the file.   Each
			unique key generates a column, and the values
			are printed in the data section of the FIAT file.
			Note that the TTYPE header lines will be automatically
			generated from ldata.
			Hdr is a dictionary of information that will be
			put in the header.
			Comments is a list of comment lines for the header.
			Fd is a file descriptor where the data should be
			written.
			Sep is a string used to separate data columns.
			Blank is a string to use when a data value is missing.
		"""
	w = writer(fd, sep=sep, blank=blank)
	if comments is not None:
		for com in comments:
			w.comment(com)
	if hdr is not None:
		w.headers(hdr)
	for d in ldata:
		w.datum(d)


class writer(gpk_writer.writer):
	__doc__ = """Write a file in FIAT format.
			Ldata is a list of dictionaries.  Each dictionary
			corresponds to one line in the file.   Each
			unique key generates a column, and the values
			are printed in the data section of the FIAT file.
			Note that the TTYPE header lines will be automatically
			generated from ldata.
			Hdr is a dictionary of information that will be
			put in the header.
			Comments is a list of comment lines for the header.
			Fd is a file descriptor where the data should be
			written.
			Sep is a string used to separate data columns.
			Blank is a string to use when a data value is missing.
			"""

	def comment(self, comment):
		"""Add a comment to the data file."""
		self.fd.write("# %s\n" % comment)


	def header(self, k, v):
		if _autogen.match(k):
			die.warn("Hdr specifies information that is automatically generated: %s" % k)
		elif _drop.match(k):
			return
		self._write_header(k, v)
	

	def __init__(self, fd, sep='\t', blank='%na'):
		gpk_writer.writer.__init__(self, fd)
		self.enc = _encoder(sep)
		self.blank = blank
		self.sep = sep
		self.map = {}
		self.columns = []
		fd.write("# fiat 1.1\n")
		self._write_header( 'COL_EMPTY', self.blank)
		self._write_header('COL_SEPARATOR', 
		 		' '.join([str(ord(sc)) for sc in self.sep])
				)

	def add_cols(self, colnames):
		n = len(self.map)
		for c in colnames:
			self.map[c] = n
			self.columns.append( c )
			self.fd.write("# TTYPE%d = %s\n" % (n+1, c))
			n += 1
	
	def _hline(self, o):
		"""o is not used, except to help set the width of each field.
		"""
		ostart = 0
		hstart = 1	# The comment symbol.
		ls = len(self.sep)
		hline = []
		for (cn, val) in zip(self.columns, o):
			w = max(1, len(val) + (ostart - hstart)/2)
			ostart += len(val) + ls
			cn = str(cn)
			hstart += len(cn) + ls
			hline.append(cn.center(w))
		self.fd.write('#' + self.sep.join(hline) + '\n')


	def _write_header(self, k, v):
		self.fd.write("# %s = %s\n" % (k, v))


	def datum(self, data_item):
		o = [ self.blank ] * len(self.map)
		# o = [ self.blank for q in self.map.keys() ]
		try:
			# This is the path for most calls.
			for (k, v) in data_item.iteritems():
				o[self.map[k]] = self.enc.fwd(str(v))
		except KeyError:
			# This is the path the first time, when self.map
			# doesn't yet exist.
			add = []
			for (k, v) in data_item.iteritems():
				if isinstance(k, types.StringType):
					pass
				elif isinstance(k, types.IntType) and k>=0:
					pass
				else:
					raise TypeError, ("Key is not a string or non-negative integer", k)
				if not k in self.map:
					# add.append( (k, len(str(v))) )
					add.append( k )
			add.sort(col_order)
			# self.add_cols([ t[0] for t in add ])
			self.add_cols( add )
			o = [ self.blank ] * len(self.map)
			# o = [ self.blank for q in self.map.keys() ]
			for (k, v) in data_item.iteritems():
				o[self.map[k]] = self.enc.fwd(str(v))
			self._hline(o)
		self.fd.write(self.sep.join(o) + '\n')


	def datavec(self, vector, numeric=False):
		"""This assumes that you've already called add_cols() to set the
		column names.   It is an error to have a vector whose length doesn't
		match the number of column names.
		"""
		lv = len(vector)
		lc = len(self.columns)
		assert lv >= lc, "vector length=%d but %d columns" % (lv, lc)
		if lc < lv:
			self.add_cols( [ '%d' % q for q in range(lc, lv) ] )
		if numeric:
			self.fd.write( self.sep.join([str(q) for q in vector]) + '\n' )
		else:
			self.fd.write( self.sep.join([self.enc.fwd(str(q)) for q in vector]) + '\n' )


class merged_writer(writer):
	"""Assumes that the data will be read with read_merged(), so that
	header values will supply default values for each column.
	"""

	def __init__(self, fd, sep='\t', blank='%na'):
		writer.__init__(self, fd, sep, blank)
		self._hdr = {}
	
	def header(self, k, v):
		self._hdr[k] = v
		writer.header(self, k, v)

	def datum(self, data_item):
		"""Assumes that the data will be read with read_merged(), so that
		header values will supply default values for each column.
		This writes a line in the fiat file, but first it deletes any values
		that alread exist as a header item of the same name.
		"""
		tmp = {}
		for (k, v) in data_item.items():
			if k not in self._hdr or v!=self._hdr[k]:
				tmp[k] = v
		writer.datum(self, tmp)




class FiatError(RuntimeError):
	def __init__(self, s):
		RuntimeError.__init__(self, s)


class ConflictingColumnSpecification(FiatError):
	def __init__(self, s):
		FiatError.__init__(self, s)


def dequote(s):
	"""Remove quotes from a value."""
	if len(s) < 2:
		return s
	elif s[0] in '\'"' and s[0]==s[-1]:
		return s[1:-1]
	return s


def _check_last_comment(comment, names):
	"""Check to see if the last comment is just a list of column names.
	This is what write() produces.   If so, it can be safely deleted.
	"""
	# print "SORTED NAMES=", sorted_names
	# print "LAST COMMENT", comment.split()
	return comment.split() == names



_encoder_cache = {}
def _encoder(sep):
	if not sep in _encoder_cache:
		if len(_encoder_cache) > 30:
			_encoder_cache.pop()
		notallowed = ' \t%\n\r'
		if sep is not None and sep not in notallowed:
			notallowed += sep
		_encoder_cache[sep] = g_encode.encoder(notallowed=notallowed)
	return _encoder_cache[sep]


class rheader(object):
	LTTYPE = len('TTYPE')

	def __init__(self):
		self.sep = None
		self.blank = '%na'
		self.comments = []
		self.name_to_col = {}
		self.header = {}
		# self.header = {'__COLUMNS': {}, '__NAME_TO_COL': {}
				# }
		self.enc = _encoder('')
		self.icol = []
		self.colmap = {}
	
	def parse(self, s):
		l = s[1:].strip()
		a = l.split('=', 1)
		if len(a) > 1 and len(a[0].split()) == 1:
			attr = a[0].strip()
			val = dequote(a[1].strip())
			if attr.upper().startswith('TTYPE'):
				ic = int(attr[self.LTTYPE:])-1
				if ic in self.colmap and self.colmap[ic]!=val:
					raise ConflictingColumnSpecification, 'column=%d: "%s" vs. "%s"' % (
										ic, val, self.icol[ic]
										)
				if val in self.name_to_col and self.name_to_col[val] != ic:
					raise ConflictingColumnSpecification, 'val="%s": columns %d vs. %d' % (
										val, ic, self.icol[ic]
										)
				self.extend_icol(ic+1)
				self.icol[ic] = val
				self.colmap[ic] = val
				self.name_to_col[val] = ic
			elif attr == 'COL_EMPTY':
				self.blank = val
			elif attr == 'COL_SEPARATOR':
				self.sep = ''.join( [chr(int(q)) for q in val.split() ] )
				self.enc = _encoder(self.sep)
			else:
				self.header[attr] = val
		elif not _check_last_comment(l, self.icol):
			self.comments.append(l)


	def extend_icol(self, la):
		if len(self.icol) < la:
			self.icol.extend( range(len(self.icol), la) )

	def dump(self, d):
		hdr = self.header
		com = self.comments
		self.header = {}
		self.comments = []
		return (hdr, d, com)

	def dumpx(self, d):
		hdr = self.header
		hdr['__NAME_TO_COL'] = self.name_to_col
		hdr['__COLUMNS'] = self.colmap
		com = self.comment
		self.header = {}
		self.comment = []
		return (hdr, d, com)

def read(fd):
	"""Read in a fiat file. Return (header, data, comments),
	where header is a dictionary of header information,
	data is a list of dictionaries, one for each line in the file,
	and comments is a list of strings.
	Each line in the FIAT file is represented by a dictionary that maps
	column name into the data (data is a string).
	Lines without data in a certain column will not have an entry in that line's
	dictionary for that column name.
	Two special entries are added to header: __COLUMNS points to a mapping
	from column numbers (the order in which they appeared in the file,
	left to right, starting with 0) to names,
	and __NAME_TO_COL is the reverse mapping.

	Note that this routine requires all header lines to precede all data
	lines.    Fiat does not make that requirement.
	"""
	out = []
	hdr = {}
	comments = []
	for (h, d, c) in readiter(fd):
		hdr.update(h)
		comments.extend(c)
		if d is not None:
			out.append(d)
	return (hdr, out, comments)


def read_merged(fd):
	"""Read in a fiat file. Returns (among other things)
	a list of dictionaries, one for each line in the file.
	Each line in the input FIAT file is represented by a dictionary that maps
	column name into the data (data is a string).
	The header data in the FIAT file is merged into the per-column data,
	so that the header data is used as a default value for a column.
	In fact, each header value creates a column of the same name, so that all
	the information in the file is in the resulting list of dictionaries.

	That this routine does not require header lines to precede data lines.
	If header lines appear in the middle, then a new column will be created from
	that point onwards.
	@return (header, data, comments)
	@rtype (dict, list(dict), list(str))
	@param fd: Where the data comes from.
	@type fd: file
	"""
	out = []
	comments = []
	hdr = {}
	for (h, d, c) in readiter(fd):
		comments.extend(c)
		if d is not None:
			hdr.update(h)
			tmp = hdr.copy()
			tmp.update(d)
			out.append(tmp)
	return (h, out, comments)



def readiter(fd):
	"""Read in a fiat file.
	Each line in the file is represented by a dictionary that maps
	column name into the data (data is a string).
	Lines without data in a certain column will not have an entry in that line's
	dictionary for that column name.
	A fiat file is a mixed header/data file.   Typically, all
	the header info is at the top, but it can be intermixed in with
	the data.    (Lines beginning with '#' are either header data
	or comments.)

	@param fd: A file descriptor to read.
	@type fd: file object
	@return: (header, data, comments),
		where header is the collected dictionary of header information
		since the last iteration,
		data is a dictionary of the data on the current line,
		and comments is a list of comment string seen so far.
		The end of the file yields None for the last data,
		along with any header info or comments after the last datum.
	@rtype: (dict(str:str), dict(str:str), list(str))
	"""
	hobj = rheader()

	line = fd.readline()
	if line.startswith('# fiat'):
		line = fd.readline()
	while line:
		line = line.rstrip('\r\n')

		if line.startswith('#'):
			hobj.parse(line)
		elif not line:		# empty line
			yield hobj.dump({})
		else:
			a = line.split(hobj.sep)
			if len(hobj.icol) < len(a):
				hobj.extend_icol(len(a))
			tmp = {}
			for (ic, ai) in zip(hobj.icol, a):
				if ai != hobj.blank:
					tmp[ic] = hobj.enc.back(ai)
			yield hobj.dump(tmp)
		line = fd.readline()
	yield hobj.dump(None)



def read_as_float_array(fd, loose=False, baddata=None):
	"""Read in a fiat file. Return (header, data, comments),
	where header is a dictionary of header information,
	data is a numpy array, and comments is a list of strings.
	Two special entries are added to header: __COLUMNS points to a mapping
	from column numbers (the order in which they appeared in the file,
	left to right, starting with 0) to names, and
	_NAME_TO_COL holds the reverse mapping.

	Empty values are set to NaN.

	If loose==False, then all entries must be either floating point numbers
	or empty entries or equal to baddata (as a string, before conversion
	to a float).
	If loose==True, all non-floats will simply be masked out.
	"""
	import Num
	import fpconst

	if loose:
		def float_or_NaN(s):
			if s == baddata:
				tmpd = NaN
			else:
				try:
					tmpd = float(s)
				except ValueError:
					tmpd = NaN
			return tmpd
	else:
		def float_or_NaN(s):
			if s == baddata:
				tmpd = NaN
			else:
				tmpd = float(s)
			return tmpd



	NaN = fpconst.NaN
	hobj = rheader()

	data = []
	maxcols = None
	line = fd.readline()
	if line.startswith('# fiat'):
		line = fd.readline()
	while line:
		line = line.rstrip('\r\n')

		if line.startswith('#'):
			hobj.parse(line)
		else:
			a = line.split(hobj.sep)
			if len(hobj.icol) < len(a):
				hobj.extend_icol(len(a))
			tmpd = Num.array( [ float_or_NaN(q) for q in a ], Num.Float)
			data.append(tmpd)
		line = fd.readline()

	# Now, if the length has expanded between the beginning and the end,
	# go back and fill out the data with NaNs to a uniform length.
	if data and data[0].shape[0] < data[-1].shape[0]:
		NaNs = Num.zeros((maxcols,), Num.Float) + NaN
		assert not Num.sometrue(Num.equal(NaNs, NaNs))
	i = 0
	while i<len(data) and data[i].shape[0]<maxcols:
		data[i] = Num.concatenate((data[i], NaNs[:maxcols-data[i].shape[0]]))
		i += 1

	return (hobj.header, data, hobj.comments)



def _hcheck(a, b):
	# del a['__NAME_TO_COL']
	# del a['__COLUMNS']
	assert abs(float(a['SAMPRATE'])-float(b['SAMPRATE'])) < 1e-8
	del a['SAMPRATE']
	del b['SAMPRATE']
	assert a == b


def _dcheck(a, b):
	for (ax, bx) in zip(a, b):
		tmp = {}
		for (k, v) in bx.items():
			tmp[k] = str(v)
		assert ax == tmp


def test1():
	fd = open("/tmp/fakeZZZ.fiat", "w")
	data = [{'a':1, 'b':2},
			{'a':2, 'b':3},
			{'b':3},
			{'a':3},
			{},
			{'a':1, 'b':0}
		]
	comments = ['Comment1', 'Comment2']
	header = {'SAMPRATE': 2.3, 'DATE':'2001-09-21T21:32:32'}
	write(fd, data, comments, header)
	fd.close()
	fd = open("/tmp/fakeZZZ.fiat", "r")
	h, d, c = read(fd)
	_hcheck(h, header)
	_dcheck(d, data)
	assert c == comments


def test2():
	fd = open("/tmp/fakeZZZ.fiat", "w")
	data = [{}, {'a': 111},
			{'a':1, 'b':2},
			{'a':2, 'b':3},
			{'b':3},
			{'a':3},
			{},
			{'a':1, 'b':0}
		]
	comments = ['Comment1', 'Comment2']
	header = {'SAMPRATE': 2.3, 'DATE':'2001-09-21T21:32:32'}
	write(fd, data, comments, header)
	fd.close()
	fd = open("/tmp/fakeZZZ.fiat", "r")
	h, d, c = read(fd)
	_hcheck(h, header)
	_dcheck(d, data)
	assert c == comments



def test3():
	fd = open("/tmp/fakeZZ1.fiat", "w")
	comments = ['Comment1', 'Comment2']
	header = {'SAMPRATE': 2.3, 'DATE':'2001/09/21'}
	data = [
		{'a':101, 'bljljlj':2, 'fd': 33341, 'q': 12},
		{'a':10, 'bljljlj':2, 'fd': 3334111, 'q': 12},
		{'a':10, 'bljljlj':21, 'fd': 3331, 'q': 12},
		{'a':1, 'bljljlj':4, 'fd': 3334122, 'q': 12},
		{'a':1, 'bljljlj':3, 'fd': 333, 'q': 12}
		]
	write(fd, data, comments, header,
		blank = 'NA'
		)
	fd.close()
	fd = open("/tmp/fakeZZ1.fiat", "r")
	h, d, c = read(fd)
	_hcheck(h, header)
	_dcheck(d, data)
	assert c == comments


def test4():
	import Num
	fd = open("/tmp/fakeZZ1.fiat", "w")
	adata = Num.zeros((4,7), Num.Float)
	for i in range(adata.shape[0]):
		for j in range(adata.shape[1]):
			adata[i,j] = i**2 + 2*j**2 - 0.413*i*j - 0.112*float(i+1)/float(j+2)
	comments = ['C1']
	hdr = {'foo': 'bar'}
	columns = ['A', 'b', 'C', 'd']
	write_array(fd, adata, columns=columns,
			comments=comments, hdr=hdr, sep='\t')
	fd = open("/tmp/fakeZZ1.fiat", "r")
	h, adtest, c = read_as_float_array(fd, loose=False, baddata=None)
	assert c == comments
	for (k,v) in hdr.items():
		assert h[k] == v
	# for (i,cname) in enumerate(columns):
		# assert h['__COLUMNS'][i] == cname
		# assert h['__NAME_TO_COL'][cname] == i
	if Num.sum(Num.absolute(adtest-adata)) > 0.001:
		die.die('Bad array recovery')

def test():
	test1()
	test2()
	test3()
	test4()

if __name__ == '__main__':
	test()


__author__ = '$Author: gpk $'
__version__ = '$Revision: 1.53 $'
__date__ = '$Date: 2007/06/30 13:31:21 $'




"""Suppose there is a random variable with true distribution p.
Then (as we will see) we could represent that random variable with a code that has average length H(p).
However, due to incomplete information we do not know p;
instead we assume that the distribution of the random variable is q.
Then (as we will see) the code would need more bits to represent the random variable.
The difference in the number of bits is denoted as D(p|q).
The quantity D(p|q) comes up often enough that it has a name: it is known as the relative entropy::

	\begin{definition}
	The {\bf relative entropy} or {\bf Kullback-Leibler distance between two
	probability mass functions p(x) and q(x) is defined as
	D(p||q) = \Sum_{x\in X} p(x) \log\frac{p(x)}{q(x)}
	\end{displaymath}
	\end{definition}

Note that this is not symmetric, and the q (the second argument) appears only in the denominator.
"""

import Num
import mcmc
import mcmc_helper
import gpkavg
import gpkmisc
import math
import random

def P(x):
	assert Num.alltrue(x >= 0)
	sx = Num.sum(x, axis=0)
	return x/sx


def multinomial_logp(x, cF):
	c, F = cF
	assert x.shape == c.shape
	if Num.sometrue(x <= 0):
		return None
	sx = Num.sum(x, axis=0)
	xe = x / sx
	log_combinations = gpkmisc.log_factorial(Num.sum(c, axis=0))
	for cc in c:
		log_combinations -= gpkmisc.log_factorial(cc)
	normc = x.shape[0] * math.log(sx)**2
	return (Num.sum(Num.log(xe)*c, axis=0) + log_combinations - normc) * F


def multinomial_fixer(x, c):
	EPS = 1e-8
	while Num.sometrue( x <= 0.0):
		for i in Num.nonzero( x <= 0.0):
			x[i] = -x[i] + EPS
	return x


def kl_nonzero_probs(p, q):
	"""Kullback-Lieber distance between two normalized,
	nonzero probability distributions."""
	# print 'kl_nonzero_probs:', p, q
	rv = Num.sum( kl_nonzero_prob_m(p, q), axis=0 )
	# print '     result=', rv
	return rv


def kl_nonzero_prob_m(p, q):
	rv = p*Num.log(p/q) / math.log(2.0)
	return rv


def kldist_vec(p, q, N=None, Fp=1.0, Fq=1.0, Clip=0.01):
	"""Relative entropy or Kullback-Liebler distance between
	two frequency distributions p and q.
	Here, we assume that both p and q are counts
	derived from multinomial distributed data;
	they are not normalized to one.
	"""

	p = Num.asarray(p, Num.Int)
	q = Num.asarray(q, Num.Int)
	assert p.shape == q.shape
	if N is None:
		N = p.shape[0]**2 * 30
	assert Num.sum(p) > 0
	assert Num.sum(q) > 0
	qstart = ((0.5+q)/Num.sum(0.5+q, axis=0))
	pstart = ((0.5+p)/Num.sum(0.5+p, axis=0))
	pV = 0.1*Num.identity(p.shape[0])/float(p.shape[0])**1.5
	qV = 0.1*Num.identity(q.shape[0])/float(q.shape[0])**1.5
	xq = mcmc.bootstepper(multinomial_logp, qstart, qV,
				c=(q,Fq), fixer=multinomial_fixer)
	xp = mcmc.bootstepper(multinomial_logp, pstart, pV,
				c=(p,Fp), fixer=multinomial_fixer)
	mcmchp = mcmc_helper.stepper(xp)
	mcmchq = mcmc_helper.stepper(xq)
	mcmchp.run_to_bottom()
	mcmchp.run_to_ergodic(5.0)
	mcmchq.run_to_bottom()
	mcmchq.run_to_ergodic(5.0)
	o = []
	for i in range(N):
		mcmchp.run_to_ergodic(1.0)
		mcmchq.run_to_ergodic(1.0)
		o.append( kl_nonzero_probs(P(xp.prms()), P(xq.prms())) )
	avg, sigma = gpkavg.avg(o, None, Clip)
	return (avg, sigma)


def tr_from_obs(x):
	"""Given a matrix of P(i and j) as x[i,j],
	we compute P(j given i) and return it as y[i,j].
	The result is a transition probability matrix
	where the first index is the input state, and the second
	index marks the result state.
	"""
	# print 'tr_from_obs:', x
	assert x.shape[0] == x.shape[1]
	y = Num.zeros(x.shape, Num.Float)
	for i in range( x.shape[0] ):
		sxi = Num.sum( x[i], axis=0 )
		assert sxi > 0
		y[i] = x[i] / sxi
	# print '\ttr_from_obs:', y
	return y


def estimate_tr_probs(counts, N, F=1.0):
	n = counts.shape[0]
	assert counts.shape == (n,n)
	q = Num.ravel(counts)
	assert Num.alltrue(Num.equal(Num.reshape(q, (n,n)), counts))
	assert Num.sum(q) > 0
	qstart = ((0.5+q)/Num.sum(0.5+q, axis=0))
	qV = 0.1*Num.identity(q.shape[0])/float(q.shape[0])**1.5
	xq = mcmc.bootstepper(multinomial_logp, qstart, qV,
				c=(q,F), fixer=multinomial_fixer)
	mcmch = mcmc_helper.stepper(xq)
	mcmch.run_to_bottom()
	mcmch.run_to_ergodic(5.0)
	o = []
	for i in range(N):
		mcmch.run_to_ergodic(1.0)
		# print 'xq.prms=', xq.prms()
		est_p_obs = Num.reshape(xq.prms(), (n,n)) / Num.sum( xq.prms(), axis=0 )
		tmptr = tr_from_obs( est_p_obs )
		# print '# tmptr=', tmptr
		o.append( tmptr )
	return o


class NoConvergenceError(ValueError):
	def __init__(self, s):
		ValueError.__init__(self, s)

class NotMarkovError(ValueError):
	def __init__(self, s):
		ValueError.__init__(self, s)

def solve_for_pi(p):
	"""Given a transition probability matrix p,
	where the first index is the initial state
	and the second index is the resultant state,
	compute the steady-state probability distribution,
	assuming a Markov process.
	"""
	MAXITERS = 50
	# print 'Solve for pi:', p
	n = p.shape[0]
	assert p.shape == (n,n)
	pi = Num.ones((n,), Num.Float)/float(n)
	nxtnorm = None
	iter = 0
	npi = None
	while iter < MAXITERS:
		nxt = Num.matrixmultiply(pi, p)
		# print "pi*p=", nxt
		nxtnorm = Num.sum(nxt, axis=0)
		# print "nxtnorm=", nxtnorm
		npi = nxt / nxtnorm
		if Num.sum((pi-npi)**2) < 1e-8:
			break
		pi = npi
		iter += 1
	if iter == MAXITERS:
		raise NoConvergenceError, (nxtnorm, Num.sum((pi-npi)**2))
	if abs(nxtnorm-1) > 1e-3:
		raise NotMarkovError, nxtnorm
	return npi


def kl_nonzero_tr_probs(pp, qq):
	"""KL distance, given a matrix of nonzero transition probabilities.
	Each matrix indexes states as pp[from,to], and contains
	P(to given from) as a conditional probability,
	where for any from, the Sum over to( pp[from,to]) = 1.
	"""
	EPS = 1e-6
	pi = solve_for_pi(pp)
	# print "pi=", pi
	o = 0.0
	for (i, pi_i) in enumerate(pi):
		if pi_i > EPS:
			# These following two normalizations are formally
			# unnecessary.
			# print 'Nspp,qq=', Num.sum(pp[i,:]), Num.sum(qq[i,:])
			assert abs( Num.sum(pp[i,:]) - 1) < 1e-4
			assert abs( Num.sum(qq[i,:]) - 1) < 1e-4
			d = kl_nonzero_probs(pp[i,:], qq[i,:])
			# print 'i,d,pi=', i, d, pi_i
			o += d * pi_i
	return o


def kl_nonzero_tr_prob_m(pp, qq):
	assert len(pp.shape) == 2
	assert pp.shape == qq.shape
	EPS = 1e-6
	pi = solve_for_pi(pp)
	o = Num.zeros(pp.shape, Num.Float)
	for (i, pi_i) in enumerate(pi):
		if pi_i > EPS:
			pnext = pp[i,:]/Num.sum(pp[i,:])
			qnext = qq[i,:]/Num.sum(qq[i,:])
			d = kl_nonzero_prob_m(pnext, qnext) * pi_i
			Num.add(o[i], d, o[i])
	return o


def cross(a, b):
	o = []
	for aa in a:
		for ab in b:
			o.append( (aa,ab) )
	return o


def kldist_Markov(p, q, N=None):
	"""Kullback-Liebler distance between two matricies
	of bigram counts.
	"""

	p = Num.asarray(p, Num.Int)
	q = Num.asarray(q, Num.Int)
	assert p.shape == q.shape
	if N is None:
		N = p.shape[0]**2 * 30
	rN = int(round(math.sqrt(2*N)))
	pP = estimate_tr_probs(p, rN)
	# print 'pP=', pP
	qP = estimate_tr_probs(q, rN)
	# print 'qP=', qP
	o = []
	for (pp,qq) in random.sample(cross(pP,qP), N):
		# print "pp=", pp
		# print "qq=", qq
		o.append(kl_nonzero_tr_probs(pp, qq))
	avg, sigma = gpkavg.avg(o, None, 0.0001)
	return (avg, sigma)


def kldist_Markov_m(p, q, N=None):
	"""Kullback-Liebler distance between two matrices
	of bigram counts.
	"""

	p = Num.asarray(p, Num.Int)
	q = Num.asarray(q, Num.Int)
	assert p.shape == q.shape
	if N is None:
		N = p.shape[0]**2 * 30
	rN = int(round(math.sqrt(2*N)))
	pP = estimate_tr_probs(p, rN)
	# print 'pP=', pP
	qP = estimate_tr_probs(q, rN)
	# print 'qP=', qP
	n = p.shape[0]
	assert q.shape[0] == n
	o = Num.zeros((n,n), Num.Float)
	oo = []
	for (pp,qq) in random.sample(cross(pP,qP), N):
		# print 'pp,qq=', pp, qq
		tmp = kl_nonzero_tr_prob_m(pp, qq)
		# print 'tmp=', tmp
		o += tmp
		oo.append(kl_nonzero_tr_probs(pp, qq))
	avg, sigma = gpkavg.avg(oo, None, 0.0001)
	return (o/(n*n), avg, sigma)


def kldist_Markov_mm(*p):
	"""List of Kullback-Liebler distances between all combinations
	of pairs of matrices of bigram counts.
	It returns a list of matrices of all the distances.
	Each item on the list is a sample of the distance histogram.
	"""

	# ??? Darned if I know what "N" is.
	rN = int(round(math.sqrt(2*N)))
	pP = []
	n = None
	for pp in p:
		ppa = Num.asarray(p, Num.Int)
		if n is None:
			n = p.shape[0]
		assert p.shape[0] == n
		pP.append( estimate_tr_probs(ppa, rN) )

	samples = random.sample(cross(range(rN), range(rN)), N)
	o = []
	for (x,y) in samples:
		tmp = Num.zeros((len(p), len(p)), Num.Float)
		for (i, ppl) in enumerate(pP):
			for (j, qql) in enumerate(pP):
				tmp[i,j] = kl_nonzero_probs(ppl[x], qql[y])
		o.append( tmp )
	return o



if __name__ == '__main__':
	# print kldist_vec([100,100], [1,100])
	print kldist_Markov(
		[[0,0,10],[0,0,10],[0,0,10]],
		[[20,0,0],[0,20,0],[0,0,20]]
		)



"""Run a *nix command and capture the result.
This module is thread-safe."""

import re
import g_pipe

comment = re.compile("\\s*#");

class ExecError(Exception):
	def __init__(self, *s):
		Exception.__init__(self, *s)


def getiter_raw(s, argv, input=None, perline=None, debug=False):
	"""Read a list of lines from a process.
	Input is a sequence or iterator containing strings to feed
	to the process on startup, before the first output is read.
	Perline is a sequence/iterator of strings to feed in, one at a time,
	as the process is producing data.
	Note that if input or perline is badly chosen, one can
	produce a locked loop of pipes.
	"""

	if argv is None:
		argv = [s]
	elif s is None:
		s = argv[0]

	if debug:
		print '#OPENING', s, argv
	wpipe, rpipe = g_pipe.popen2(s, argv)

	if wpipe is None or rpipe is None:
		raise ExecError('Cannot spawn pipe for {%s}' % s, *argv)
	if input is not None:
		if debug:
			for t in input:
				print '#WRITING <%s>' % t
				wpipe.write(t)
		else:
			for t in input:
				wpipe.write(t)

	if perline is None:
		wpipe.close()
		wpipe = None
	else:
		wpipe.flush()
		perline = perline.__iter__()

	while True:
		if wpipe is not None:
			try:
				wpipe.write(perline.next())
				wpipe.flush()
			except StopIteration:
				wpipe.close()
				wpipe = None
				pass
		line = rpipe.readline();
		if not line:
			break
		if debug:
			print '#LINE:', line.rstrip('\n')
		yield line

	if wpipe is not None:
		wpipe.close()
		if perline is not None:
			raise ExecError("Unused input: program {%s} terminated before perline ran out of data" % s,
					*argv)
	sts = rpipe.close()
	# if sts is None:
		# sts = 0
	if sts != 0 :
		raise ExecError('spawned command fails with %d from {%s}' % (sts, s), *argv)


def getiter(s, argv, input=None, perline=None, debug=False):
	"""Read a list of lines from a process, after
	dropping junk like comments (#).
	Returns a trouble indication if the process returns
	'ERR:'
	Input is a sequence or iterator containing strings to feed
	to the process on startup, before the first output is read.
	Perline is a sequence/iterator of strings to feed in, one at a time,
	as the process is producing data.
	Note that if input or perline is badly chosen, one can
	produce a locked loop of pipes.
	"""

	for line in getiter_raw(s, argv, input=input, perline=perline, debug=debug):
		if line.startswith('#'):
			continue;
		line = line.rstrip()
		if line.startswith('ERR:'):
			raise ExecError('%s from {%s}' % (line, s), *argv)
		cs = comment.search(line)
		if cs is not None:
			line = cs.string[:cs.start()]
		yield line


def get(s, argv=None, input=None, perline=None, debug=False):
	"""Read a single line from a process, after
	dropping junk like comments (#).
	Raises an exception if the process returns
	'ERR:' or produces no output.
	"""
	try:
		return getiter(s, argv, input, perline=perline, debug=debug).next()
	except StopIteration:
		raise ExecError('no output fron {%s}' % s, *argv)


def get_raw(s, argv=None, input=None, perline=False, debug=False):
	"""Read a single line from a process.
	Returns None if there is no output.
	"""
	try:
		return getiter(s, argv, input, perline=perline, debug=debug).next()
	except StopIteration:
		return None

def getlast(s, argv=None, input=None, perline=False, debug=False):
	"""Read a single line from a process, after
	dropping junk like comments (#).
	Raises an exception if the process returns
	'ERR:' or produces no output.
	"""
	ok = False
	for q in getiter(s, argv, input, perline=perline, debug=debug):
		ok = True
	if ok:
		return q
	else:
		raise ExecError('no output fron {%s}' % s, *argv)

def getall(s, argv, input=None, perline=None, debug=False):
	return [ q for q in getiter(s, argv, input=input,
					perline=perline, debug=debug
				)
		]



def test():
	inp = ['once\n', 'upon\n', 'a\n', 'time.\n']
	die.note('inp', inp)
	oup = getall(None, ['cat'], input=inp)
	die.note('oup', oup)
	if inp != ["%s\n" % q for q in oup]:
		die.die('whoops: bad cat')
	

if __name__ == '__main__':
	import die
	test()
	test()
	test()
	print 'OK'



"""This is a helper module to make use of mcmc.py and mcmc_big.py.
It allows you to conveniently run a Monte-Carlo simulation of any
kind until it converges (L{stepper.run_to_bottom}) or until it
has explored a large chunk of parameter space (L{stepper.run_to_ergodic}).

It also helps you with logging the process.
"""

import sys
from mpi4py import MPI		# This uses mpi4py
import random
from gmisclib import die
from gmisclib import Num
from gmisclib import mcmc
from gmisclib import mcmc_helper as MCH
MCH.Debug = True
from gmisclib.mcmc_helper import TooManyLoops, warnevery, logger_template, test

mpi = MPI.COMM_WORLD

class _counter(object):
	def __init__(self):
		self.count = 0

	def update(self, accepted):
		self.count += accepted

	def get(self):
		return self.count



class _accum_erg(object):
	def __init__(self):
		self.erg = 0.0
		self.last_resetid = -1

	def update(self, x):
		"""Increment an accumulator, but reset it to
		zero if the underlying stepper has called reset().
		The increment is by the ergodicity estimate, so
		C{self.erg} should approximate how many times the
		MCMC process had wandered across the probability
		distribution.
		"""
		tmp = x.reset_id()
		if tmp != self.last_resetid:
			self.erg = 0.0
			self.last_resetid = tmp
		else:
			self.erg += x.ergodic()


	def get(self):
		return self.erg


class stepper(MCH.stepper):
	def __init__(self, x, maxloops=-1, logger=None):
		die.info('# mpi stepper rank=%d size=%d' % (mpi.rank, mpi.size))
		assert maxloops == -1
		self.maxloops = -1
		MCH.stepper.__init__(self, x, maxloops, logger)


	def reset_loops(self, maxloops=-1):
		assert maxloops == -1
		self.maxloops = -1


	def communicate_hook(self):
		self.note('chook iter=%d' % self.iter)
		if mpi.size > 1:
			self.note('chook active iter=%d' % self.iter)
			c = self.x.current()
			v = c.vec()
			lp = c.logp()

			if mpi.rank%2 == 0:
				mpi.Send((v, lp), (mpi.rank+1)%mpi.size, tag=self.MPIID)
			nv, nlp = mpi.Recv(tag=self.MPIID)
			if mpi.rank%2 == 1:
				mpi.Send((v, lp), (mpi.rank+1)%mpi.size, tag=self.MPIID)

			if nlp > lp - self.x.T*random.expovariate(1.0):
				self.x._set_current(c.new(nv-v, logp=nlp))
				self.note('communicate succeeded')
			else:
				self.note('communicate not accepted')
			sys.stdout.flush()


	MPIID = 1241

	# def _not_enough_changes(self, n, ncw):
		# mytmp = n < ncw
		# self.note('_nyc pre')
		# ntrue = mpi.Allreduce(int(mytmp), MPI.SUM)
		# self.note('_nyc post')
		# return ntrue*3 >= mpi.size


	def _not_yet_ergodic(self, nc, ncw):
		mytmp = nc.get() < ncw
		self.note('_nye pre')
		ntrue = mpi.Allreduce(int(mytmp), MPI.SUM)
		self.note('_nye post')
		return ntrue*2 >= mpi.size


	def _not_at_bottom(self, xchanged, nchg, es, dotchanged, ndot, update_T):
		mytmp = (Num.sometrue(Num.less(xchanged,nchg))
					or es<1.0 or dotchanged<ndot
					or (update_T and self.x.T>1.5)
				)
		self.note('_nab pre')
		ntrue = mpi.Allreduce(int(mytmp), MPI.SUM)
		self.note('_nab post')
		return ntrue*4 >= mpi.size


	def join(self, id):
		self.note('pre join %s' % id)
		rootid = mpi.Bcast(id)
		assert rootid == id
		self.note('post join %s' % id)


	def note(self, s):
		if MCH.Debug:
			print s, "rank=", mpi.rank
			sys.stdout.flush()


def precompute_logp(lop):
	"""Does a parallel evaluation of logp for all items in lop.
	"""
	nper = len(lop)//mpi.size
	r = mpi.rank
	mychunk = lop[r*nper:(r+1)*nper]
	for p in mychunk:
		p.logp()
	for r in range(mpi.size):
		nc = mpi.Bcast(mychunk, r)
		lop[r*nper:(r+1)*nper] = nc
	mpi.Barrier(0)




def test():
	def test_logp(x, c):
		# print '#', x[0], x[1]
		return -(x[0]-x[1]**2)**2 + 0.001*x[1]**2
	x = mcmc.bootstepper(test_logp, Num.array([0.0,2.0]),
				Num.array([[1.0,0],[0,2.0]]))
	print 'rank=', mpi.rank
	thr = stepper(x)
	# nsteps = thr.run_to_bottom(x)
	# print '#nsteps', nsteps
	# assert nsteps < 100
	for i in range(2):
		print 'RTC'
		thr.run_to_change(2)
		print 'RTE'
		thr.run_to_ergodic(1.0)
		print 'DONE'
	thr.close()


if __name__ == '__main__':
	test()



"""Operations on dictionaries."""


__version__ = "$Revision: 1.9 $"

def filter(d, l):
	"""Passes through items listed in l."""
	o = {}
	for t in l:
		o[t] = d[t]
	return o


def remove(d, l):
	"""Removes through items listed in l."""
	o = d.copy()
	for t in l:
		if o.has_key(t):
			del o[t]
	return o



def rev1to1(d):
	o = {}
	for (k, v) in d.items():
		if v in o:
			raise ValueError, 'Dictionary is not 1 to 1 mapping'
		o[v] = k
	return o


def add_dol(dict, key, value):
	"""Add an entry to a dictionary of lists.
	A new list is created if necessary; else the value is
	appended to an existing list.
	Obsolete: replace with dict_of_lists.
	"""
	try:
		dict[key].append(value)
	except KeyError:
		dict[key] = [ value ]


def add_doc(dict, key, value):
	"""Add an entry to a dictionary of counters.
	A new counter is created if necessary; else the value is
	added to an existing counter.
	Obsolete: replace with dict_of_accums.
	"""
	try:
		dict[key] += value
	except KeyError:
		dict[key] = value


class dict_of_lists(dict):
	"""A dictionary of lists."""

	def add(self, key, value):
		"""Append value to the list indexed by key."""
		try:
			self[key].append(value)
		except KeyError:
			self[key] = [ value ]

	def addgroup(self, key, values):
		"""Append values to the list indexed by key."""
		try:
			self[key].extend(values)
		except KeyError:
			self[key] = values

	def add_ifdifferent(self, key, value):
		"""Append value to the list indexed by key."""
		try:
			v = self[key]
		except KeyError:
			self[key] = [value]
		else:
			if value not in v:
				v.append( value )

	def copy(self):
		"""This does a shallow copy."""
		return dict_of_lists(self)


	def merge(self, other):
		for (k, v) in other:
			self.addgroup(k, v)


class dict_of_sets(dict):
	"""A dictionary of lists."""
	def add(self, key, value):
		"""Append value to the set indexed by key."""
		try:
			self[key].add(value)
		except KeyError:
			self[key] = set([value])

	def addgroup(self, key, values):
		"""Append values to the list indexed by key."""
		try:
			self[key].update(values)
		except KeyError:
			self[key] = set(values)


	def copy(self):
		"""This does a shallow copy."""
		return dict_of_sets(self)


	def merge(self, other):
		for (k, v) in other:
			self.addgroup(k, v)


class dict_of_accums(dict):
	"""A dictionary of accumulators.
	Note that the copy() function produces a dict_of_accums."""

	def add(self, key, value):
		"""Add value to the value indexed by key."""
		try:
			self[key] += value
		except KeyError:
			self[key] = value


	def copy(self):
		return dict_of_accums(self)


	def merge(self, other):
		for (k, v) in other:
			self.add(k, v)


class dict_of_averages(dict):
	"""A dictionary of accumulators.
	Note that the copy() function produces a dict_of_averages."""

	def add(self, key, value, weight):
		"""Add value to the value indexed by key."""
		assert weight >= 0.0
		if key in self:
			self[key] += complex(value*weight, weight)
		else:
			self[key] = complex(value*weight, weight)

	def get_avg(self, key):
		vw = self[key]
		return vw.real/vw.imag

	def copy(self):
		return dict_of_averages(self)


	def merge(self, other):
		for (k, v) in other:
			self.add(k, v)


class dict_of_maxes(dict):
	"""A dictionary of maxima.   Add a new number and the stored
	maximum will increase iff the new value is bigger.
	Note that the copy() function produces a dict_of_accums."""

	def add(self, key, value):
		"""Add value to the value indexed by key."""
		try:
			tmp = self[key]
			if value > tmp:
				self[key] = value
		except KeyError:
			self[key] = value


	def copy(self):
		return dict_of_maxes(self)


	def merge(self, other):
		for (k, v) in other:
			self.add(k, v)



class dict_of_X(dict):
	"""A dictionary of arbitrary things.
	Note that the copy() function produces a dict_of_X."""

	def __init__(self, constructor, incrementer):
		self.constructor = constructor
		self.incrementer = incrementer

	def add(self, key, value):
		"""Add value to the value indexed by key."""
		try:
			self.incrementer(self[key], value)
		except KeyError:
			self[key] = self.constructor(value)


class list_of_dicts(list):
	def __init__(self, *arg):
		list.__init__(self, arg)
	
	def lookup1(self, key, value):
		for d in self:
			try:
				if d[key]==value:
					return d
			except KeyError:
				pass



import os
import re
import die


def uq(s, qc):
        if qc and len(s)>=2 and s[0]==qc and s[-1]==qc:
		return s[1:-1]  
	return s


class hdr:
	_bad = re.compile('[^a-zA-Z_0-9.-]')

	def __init__(self, fn):
		self.fn = fn
		self.data = None
		self.modified = 0


	def has_key(self, key):
		if self.data is None:
			self.read()
		return self.data.has_key(key)


	def items(self):
		if self.data is None:
			self.read()
		return self.data.items()


	def  get(self, key, default = None, qc=None):
		if self.data is None:
			self.read()
		o = self.data.get(key, default)
		if qc:
			return uq(o, qc)
		return o


	def set(self, key, value, qc=None):
		if self.data is None:
			self.read()
		if self.has_key(key):
			self.data['#%s' % key] = self.data[key]
		if qc:
			self.data[key] = '%s%s%s' % (qc, value, qc)
		else:
			self.data[key] = value
		self.modified += 1


	def __del__(self):
		self.close()


	def close(self):
		if not self.modified:
			return
		if self.data is None:
			return
		self.write()

		


	def write(self):
		fd = open('%s.tmp' % self.fn, "w")
		tmp = self.data.items()
		tmp.sort()
		for (k, v) in tmp:
			if self._bad.search(k) and not k.startswith('#'):
				die.warn("Bad key will be ignored: %s" % k)
			else:
				fd.writelines("%s=%s\n" % (k, v))
		fd.close()
		if os.access(self.fn, os.R_OK):
			os.rename(self.fn, '%s.bak' % self.fn)
		os.rename('%s.tmp' % self.fn, self.fn)
		self.modified = 0


	def read(self):
		self.data = { 'FILENAME': self.fn}
		die.note("file", self.fn)
		n = 1
		for x in open(self.fn, "r"):
			die.note("line", n)
			x = x.strip()
			if x == '':
				n += 1
				continue
			try:
				k, v = x.split('=', 1)
				self.data[k] = v
			except:
				if not x.startswith('#'):
					raise
			n += 1



if __name__ == '__main__':
	import sys
	x = hdr(sys.argv[1])
	x.set("TEST", 1, qc=1)
	x.write()




#!/usr/bin/env python

"""Handles interactions with ds9 (U{http://hea-www.harvard.edu/RD/ds9/}).
Ds9 is an image display and analysis tool commonly used in astrophysics,
but it's good for any kind of greyscale image.

Note: there is an extra, special property Coordsys.   This is not fed to
ds9.  It is merely something that you can use to remember what coordinate
system the regions are in.
"""


import re
import threading
import os
import time
import copy as COPY
from gmisclib import die
from gmisclib import g_exec

COLORS = ('white', 'black', 'red', 'green', 'blue', 'cyan', 'magenta', 'yellow')
PLOTSYMS = ('circle', 'diamond', 'plus', 'cross')

class UnknownProperty(KeyError):
	def __init__(self, *s):
		KeyError.__init__(self, *s)


class execwait(object):
	def __init__(self, s):
		self.t = threading.Thread(target=os.system, args=(s,),
						name='execwait: %s' % s
						)
		self.t.start()

	def wait(self):
		self.t.join()


class ds9_io(object):
	def __init__(self, ds9start=None, xpaname='ds9',
				xpaset=['xpaset', ],
				xpaset_p=['xpaset', '-p'],
				xpaget=['xpaget']
				):
		if ds9start is None:
			ds9start = 'ds9'
		self.setap = tuple(xpaset_p) + (xpaname,)
		self.seta = tuple(xpaset) + (xpaname,)
		self.geta = tuple(xpaget) + (xpaname,)
		self.ds9ew = None
		self.debug = 0

		try:
			tmp = self.getall('about')
			if tmp:
				self.closed = False
				return
		except g_exec.ExecError:
			pass

		self.ds9ew = execwait(ds9start)
		sleep = 0.10
		time.sleep(2*sleep)
		while True:
			try:
				tmp = self.getall('about')
				if tmp:
					self.closed = False
					return
			except g_exec.ExecError:
				sleep += 0.1
				time.sleep(sleep)
				


	def set(self, *args):
		tmp = self.setap + tuple(args)
		g_exec.getall(tmp[0], tmp)


	def set_with_input(self, input, *args):
		"""Run the X{xpaset} command and feed it data on its standard input.
		Each element in the input list becomes one line of data
		to C{xpaset}.
		@param input:
		@type input: L{list} or iterator
		@param args: The command line used to execute C{xpaset}.
		@type args: C{[ "xpaset", ...]}
		"""
		tmp = self.seta + tuple(args)
		g_exec.getall(tmp[0], tmp,
				input=[q+'\n' for q in input],
				debug=self.debug)


	def getlast(self, *args):
		return self.getall(*args)[-1]


	def getall(self, *args):
		return list(self.getiter(*args))


	def getiter(self, *args):
		tmp = self.geta + tuple(args)
		for q in g_exec.getiter_raw(tmp[0], tmp):
			yield q.rstrip()


	def close(self):
		if self.ds9ew and not self.closed:
			self.set('quit')
			self.ds9ew.wait()
			self.ds9ew = None
			self.closed = True


	def __del__(self):
		try:
			self.close()
		except g_exec.ExecError:
			# Too late to report an error now.
			# If it doesn't shut down, that's just too bad.
			pass


class BadTagError(ValueError):
	def __init__(self, *args):
		ValueError.__init__(self, *args)


class Region(object):
	def __init__(self, kwargs, coordsys):
		self._props = kwargs
		self._pshared = True
		self.Coordsys = coordsys

	def __str__(self):
		raise RuntimeError, 'Virtual function'
	
	def addprop(self, k, v):
		if self._pshared:
			self._props = self._props.copy()
			self._pshared = False
		self._props[k] = v

	def has_prop(self, k):
		return self._props.has_key(k)

	def getprop(self, *args):
		return self._props.get(*args)


	def props(self):
		o = [ '#' ]
		for (k, v) in self._props.items():
			try:
				pof = PROPout[k]
			except KeyError, x:
				raise UnknownProperty( *(x.args) )
			o.extend( pof(k, v) )
		if len(o) > 1:
			return ' '.join(o)
		return ''


	def chgprops(self, **kwargs):
		if self._pshared:
			self._props = self._props.copy()
			self._pshared = False
		self._props.update(kwargs)
		return self

	def compatible_coord(self, coordsys):
		return (self.Coordsys is None or coordsys is None
			or self.Coordsys == coordsys)



def copy(r, **kwargs):
	"""Copy a region and update some of the keyword
	arguments (e.g. color).   The original object is unchanged,
	and you can copy any subclass of Region.
	"""
	assert isinstance(r, Region)
	tmp = COPY.copy(r)	# Deepcopy?
	tmp.chgprops(**kwargs)
	return tmp


class line(Region):
	def __init__(self, x0, y0, x1, y1, Coordsys=None, **kwargs):
		Region.__init__(self, kwargs, Coordsys)
		self.x0 = x0
		self.y0 = y0
		self.x1 = x1
		self.y1 = y1

	def __str__(self):
		return 'line %10g %10g %10g %10g %s' % ( self.x0+1, self.y0+1,
						self.x1+1, self.y1+1, self.props() )

	def end(self, whichend):
		if whichend == 0:
			return point(self.x0, self.y0, **self._props)
		elif whichend == 1:
			return point(self.x1, self.y1, **self._props)
		else:
			raise ValueError, "Whichend == 0 or 1"

	def center(self):
		return point(0.5*(self.x0+self.x1), 0.5*(self.y0+self.y1), Coordsys=self.Coordsys)


class circle(Region):
	def __init__(self, x, y, r, Coordsys=None, **kwargs):
		Region.__init__(self, kwargs, Coordsys)
		self.x = x
		self.y = y
		self.r = r

	def __str__(self):
		return 'circle %10g %10g %10g %s' % ( self.x+1, self.y+1, self.r, self.props() )

	def center(self):
		return point(self.x, self.y, Coordsys=self.Coordsys)


class point(Region):
	def __init__(self, x, y, Coordsys=None, **kwargs):
		Region.__init__(self, kwargs, Coordsys)
		self.x = x
		self.y = y

	def __str__(self):
		return 'point %10g %10g %s' % ( self.x+1, self.y+1, self.props() )




class text(Region):
	def __init__(self, x, y, text, Coordsys=None, **kwargs):
		Region.__init__(self, kwargs, Coordsys)
		self.addprop('text', text)
		self.x = x
		self.y =  y
		self.text = text

	def __str__(self):
		return 'text %10g %10g %s' % ( self.x+1, self.y+1, self.props() )


def dequote(s):
	if s.startswith('{') and s.endswith('}'):
		return s[1:-1]
	elif len(s)>1 and s.startswith('"') and s.endswith('"'):
		return s[1:-1]
	elif len(s)>1 and s.startswith("'") and s.endswith("'"):
		return s[1:-1]
	return s



PROPin = {'color': str, 'text': dequote, 'font': dequote,
	'select': int, 'edit': int, 'move': int, 'rotate': int,
	'delete': int, 'fixed': int,
	'line': None, 'tag': None,
	'ruler': int, 'width': int,
	'point': str, 'highlite': int, 'include': int
	}


def _kvquote(k, v):
	return [ '%s={%s}' % (k, v) ]


def _kvint(k, v):
	return [ '%s=%d' % (k, v) ]

def _kvstr(k, v):
	return [ '%s=%s' % (k, v) ]

def _kvnull(k, v):
	return []

def _kvline(k, v):
	return [ '%s = %d %d' % (k, v[0], v[1]) ]

_oktag = re.compile("^[a-zA-Z][a-zA-Z0-9]*$")
def _kvtags(k, v):
	for tmp in v:
		if not _oktag.match(tmp):
			raise BadTagError, 'tags can contain only alphanumerics: "%s"' % tmp
	return [ 'tag={%s}' % tmp for tmp in v ]

def _kverr(k, v):
	raise RuntimeError, "'tags', not 'tag'."

PROPout = {'color': _kvstr, 'text': _kvquote, 'font': _kvquote,
	'select': _kvint, 'edit': _kvint, 'move': _kvint, 'rotate': _kvint,
	'delete': _kvint, 'fixed': _kvint,
	'line': _kvline, 'tags': _kvtags, 'tag': _kverr,
	'ruler': _kvint, 'width': _kvint,
	'point': _kvstr, 'highlite': _kvint, 'include': _kvint
	}



_rparse = re.compile(r"""{[^}]*}|"[^"]*"|'[^']'|[^\s,()=#]+|[#]|=""")

def make_chunks(s):
	return _rparse.findall(s)



def property_parser(ap):
	kwargs = {}
	i = 0
	while i+2 < len(ap):
		arg = ap[i]
		if arg in PROPin and ap[i+1] == '=':
			if arg == 'line' and i+3 < len(ap):
				kwargs[arg] = ( int(ap[i+2]), int(ap[i+3]) )
				i += 4
			elif arg == 'tag': 
				if 'tags' not in kwargs:
					# This is just setting up a list to extend if we
					# have more than one occurrence of 'tags'.
					kwargs['tags'] = [ dequote(ap[i+2]) ]
				else:
					kwargs['tags'].append( dequote(ap[i+2]) )
				i += 3
			else:
				kwargs[arg] = PROPin[arg]( ap[i+2] )
				i += 3
		else:
			die.info('ap=%s' % str(ap))
			die.die('Unexpected word in property: %s' % arg)
	# print 'PP:', kwargs
	return kwargs


def test_property_parser():
	print "TPP"
	pp = property_parser
	assert pp(['select', '=', '1']) == {'select': 1}
	assert pp(['text', '=', '1']) == {'text': '1'}
	assert pp(['text', '=', '{foo}']) == {'text': 'foo'}
	assert pp(['text', '=', '"foo"']) == {'text': 'foo'}
	assert pp(['line', '=', '1', '1']) == {'line': (1,1)}
	assert pp(['select', '=', '1', 'fixed', '=', '0']) == {'select': 1, 'fixed': 0}


def region_parser(s, defprop):
	s = s.strip()
	kwargs = defprop.copy()
	if s.startswith('-'):
		kwargs['include'] = 0
		s = s[1:]
	elif s.startswith('+'):
		kwargs['include'] = 1
		s = s[1:]

	# print 's=', s
	aa = make_chunks(s)
	x = aa[0]
	# print 'aa=', aa
	if '#' in aa:
		hi = aa.index('#')
		# print 'hi=', hi
		ap = aa[hi+1:]
		aa = aa[1:hi]
	else:
		aa = aa[1:]
		ap = []
	# print 'region_parser:', aa, '###', ap

	args = [ float(q) for q in aa ]
	kwargs.update( property_parser(ap) )
	return (x, args, kwargs)


def test_region_parser():
	print "TRP"
	x, args, kwargs = region_parser('line(0,0,1,2)', {})
	assert x == 'line'
	assert args == [0, 0, 1, 2]
	assert kwargs == {}

	x, args, kwargs = region_parser('+point(-1,2)', {})
	assert x == 'point'
	assert args == [-1, 2]
	assert kwargs == {'include': 1}

	x, args, kwargs = region_parser('-point(-1,2) # color=red', {})
	assert x == 'point'
	assert args == [-1, 2]
	assert kwargs == {'include': 0, 'color': 'red'}

	x, args, kwargs = region_parser('circle(0,2,3) # text={foo}', {})
	assert x == 'circle'
	assert args == [0,2,3]
	assert kwargs == {'text': 'foo'}

	x, args, kwargs = region_parser('circle(0,2,3) # text="foo" rotate =1', {})
	assert x == 'circle'
	assert args == [0,2,3]
	assert kwargs == {'text': 'foo', 'rotate': 1}

	x, args, kwargs = region_parser('circle(0,2,3) # text="foo bar" rotate = 0 fixed= 1', {})
	assert x == 'circle'
	assert args == [0,2,3]
	assert kwargs == {'text': 'foo bar', 'rotate': 0, 'fixed': 1}

	x, args, kwargs = region_parser('circle(0,2,3) # text={foo bar} line=1 1 fixed=0', {})
	assert x == 'circle'
	assert args == [0,2,3]
	assert kwargs == {'text': 'foo bar', 'line': (1,1), 'fixed': 0}



_text_region = re.compile(r'\s*#\s*text[( \t]([0-9eE.+-]+)[, \t]+([0-9eE.+-]+)[) \t]\s*(.*)$')

def text_kluge(s):
	m = _text_region.match(s)
	assert m
	kwargs = property_parser(make_chunks(m.group(3)))
	# print '#kwargs=', kwargs
	txt = kwargs.pop('text')
	return text(float(m.group(1)), float(m.group(2)), txt, **kwargs)

def test_text_kluge():
	print "TTK"
	tmp = text_kluge('# text(2,1) text={fred} color=red tag={RefImgF}')
	assert tmp.getprop('color') == 'red'
	assert tmp.getprop('text') == 'fred'
	assert tmp.getprop('tags') == ['RefImgF',]
	assert abs(tmp.x-2.0)<0.0001 and abs(tmp.y-1.0)<0.00001


def r_list_factory(slist, Coordsys):
	gp = {'include': 1, 'text': '', 'color': 'green',
		'font': 'helvetica 10 normal', 'edit': 1,
		'move': 1, 'delete': 1, 'fixed': 0
		}
	for s in slist:
		s = s.strip()
		# print 's=', s
		if _text_region.match(s):
			yield text_kluge(s)
			continue
		elif s.startswith('#'):
			# print 'Comment', s
			continue
		if s.startswith('global'):
			gp = property_parser(make_chunks(s[len('global'):]))
			continue
		x, args, kwargs = region_parser(s, gp)
		if x is None:
			pass
		elif x == 'physical':
			pass
		elif x == 'image':
			pass
		elif x == 'circle':
			assert len(args) == 3
			yield circle(args[0]-1, args[1]-1, args[2], Coordsys=Coordsys, **kwargs)
		elif x == 'line':
			assert len(args) == 4
			yield line(args[0]-1, args[1]-1,
					args[2]-1, args[3]-1, Coordsys=Coordsys, **kwargs)
		elif x == 'point':
			assert len(args) == 2
			yield point(args[0]-1, args[1]-1, Coordsys=Coordsys, **kwargs)
		elif x == 'text':
			if len(args) == 2:
				# We happen to know that one of the kwargs
				# is text, which will provide the third
				# argument to the text() constructor.
				yield text(args[0]-1, args[1]-1, Coordsys=Coordsys, **kwargs)
			elif len(args) == 3:
				yield text(args[0]-1, args[1]-1, args[2], Coordsys=Coordsys, **kwargs)
		else:
			die.die('Unsupported region: %s in <%s>' % (x, s))


def read_regions_iter(fd, Coordsys=None):
	"""Read regions from a file.   This is an iterator.
	Fd is a file descriptor.
	"""
	return r_list_factory(fd, Coordsys)


def read_regions(fd, Coordsys=None):
	"""Read regions from a file.
	Fd is a file descriptor.
	"""
	return list(r_list_factory(fd, Coordsys))


PLOTSTYLES = ('discrete', 'line', 'step', 'quadratic', 'errorbar')


def _format_plot(xylist, format):
	if format == '(x,y)':
		ds9fmt = 'xy'
		o = [ '%10g %10g' % xy for xy in xylist ]
	elif format == '(x,y,ex)':
		ds9fmt = 'xyex'
		o = [ '%10g %10g %6g' % xye for xye in xylist ]
	elif format == '(x,y,ey)':
		ds9fmt = 'xyey'
		o = [ '%10g %10g %6g' % xye for xye in xylist ]
	elif format == '(x,y,ex,ey)':
		ds9fmt = 'xyexey'
		o = [ '%10g %10g %6g %6g' % xyexey for xyexey in xylist ]
	elif format == 'array(:,xy)':
		ds9fmt = 'xy'
		o = [ '%10g %10g' % (xylist[i,0], xylist[i,1]) for i in range(xylist.shape[0]) ]
	elif format == 'array(:,xyey)':
		ds9fmt = 'xyey'
		o = [ '%10g %10g %6g' % (xylist[i,0], xylist[i,1], xylist[i,2])
				for i in range(xylist.shape[0])
				]
	elif format == 'array(xy,:)':
		ds9fmt = 'xy'
		o = [ '%10g %10g' % (xylist[0,i], xylist[1,i]) for i in range(xylist.shape[1]) ]
	elif format == 'array(xyey,:)':
		ds9fmt = 'xyey'
		o = [ '%10g %10g %6g' % (xylist[0,i], xylist[1,i], xylist[2,i])
				for i in range(xylist.shape[1])
				]
		
	else:
		raise ValueError, 'Unknown plot data format: %s' % format
	return (ds9fmt, o)




class ds9(ds9_io):
	def __init__(self, ds9start=None):
		ds9_io.__init__(self, ds9start=ds9start)
		self.set('mode', 'pointer')
		self.frame = -1

	def load(self, fn):
		self.set('regions', 'delete', 'all')
		self.set('file', fn)

	def get_current_fname(self):
		return self.getlast( 'iis', 'filename' )

	def get_frames(self):
		tmp = self.getlast( 'frame', 'all' ).split()
		if 'XPA$END' in tmp:
			# This works around a bug in ds9:
			# if it has no data frames, it'll tell you
			# about the buttons and other stuff.
			return []
		return tmp

	def frame_cmd(self, *cmds):
		"""A catch-all command for anything that's not otherwise implemented."""
		self.set('frame', *cmds)
	

	def del_frames(self, framelist):
		"""Note: deleting all the frames crashes ds9.
		"""
		for i in framelist:
			self.select_frame(i)
			self.frame_cmd('delete')


	def select_frame(self, i):
		"""Select a frame on the attached ds9 image display.
		The frame will be created if it didn't already exist.
		@param i: Which frame?   Apparently, any integer is OK,
			even negative.
		@type i: L{int}
		"""
		si = str(int(i))
		if si != self.frame:
			self.set('frame', 'frameno', si)
			self.frame = si


	def get_regions_iter(self):
		return r_list_factory(self.getiter('regions', '-format', 'ds9',
							'-delim', 'nl' ),
					None
					)

	def get_regions(self):
		return list(self.get_regions_iter())

	def delete_all_regions(self):
		self.set('regions', 'delete', 'all')

	def delete_region_group(self, *group):
		for g in group:
			self.set('regions', 'group', g, 'delete')

	def add_region(self, *region):
		"""Add a list of regions to the current frame.
		@param region: one or more regions.
		@type region: C{[ region ]}, where C{region} is normally 
			an instance of L{Region}, but can be anything that
			produces a string suitable for X{ds9}.
			See U{http://hea-www.harvard.edu/RD/ds9/ref/xpa.html#regions}.
		"""
		# print 'region=', str(region)
		self.set_with_input([ str(r) for r in region ], 'regions')


	def region_cmd(self, *cmds):
		"""A catch-all command for anything that's not otherwise implemented.
		This does C{xpaset -p}, so you cannot use it for region definitions
		that need multi-line input via L{set_with_input}.
		"""
		self.set('regions', *cmds)

	def zoom(self, z):
		self.set('zoom', 'to', str(z))

	def scale_cmd(self, *cmds):
		"""A catch-all command for anything that's not otherwise implemented."""
		self.set('scale', *cmds)

	def cmap_cmd(self, *cmds):
		"""A catch-all command for anything that's not otherwise implemented."""
		self.set('cmap', *cmds)
	
	def set_mode(self, m):
		assert m in ['tile', 'blink', 'single']
		if m == 'tile':
			self.set('tile', 'yes')
		else:
			self.set('tile', 'no')
		if m == 'single':
			self.set('single')
		if m == 'blink':
			self.set('blink')

	def pan_to(self, x, y, coordsys):
		self.set('pan', 'to', str(x), str(y), coordsys)

	plotname_ok = re.compile('[a-z][a-zA-Z_]*')

	def plot(self, name, xylist, title='', xlabel='', ylabel='',
			format='(x,y)', line='solid',
			color='black', symbol=None):
		assert name != 'new'
		assert self.plotname_ok.match(name), "Bad plot name: %s" % name
		ds9fmt, formatted = _format_plot(xylist, format)
		if name in self.list_plots():
			assert title==''
			assert xlabel==''
			assert ylabel==''
			self.set_with_input( formatted, 'plot %s data %s' % (name, ds9fmt))
		else:
			self.set_with_input( formatted,
					'plot new name %s {%s} {%s} {%s} %s'
						% (name, title, xlabel, ylabel, ds9fmt)
						)
		assert color in COLORS
		self.set('plot', 'view', 'discrete', ['yes', 'no'][symbol is None])
		self.set('plot', 'view', 'line', ['yes', 'no'][line is None])
		if symbol is not None:
			assert symbol in PLOTSYMS
			self.set('plot', 'color', 'discrete', color)
			self.set('plot', 'line', 'discrete', symbol)
		if line is not None:
			self.set('plot', 'color', 'line', color)
			self.set('plot', 'line', 'dash', ['no', 'yes'][line=='solid'])

	def close_plot(self, name):
		self.set('plot', 'close', '{%s}' % name)

	def list_plots(self):
		return self.getlast('plot').split()

	def comment(self, s):
		pass

	def flush(self):
		pass


class ds9_file(object):
	"""This is just a convenient way to write regions to a file,
	instead of writing them to a ds9 session.   Note that plots
	and other interactive things are simply no-ops.
	"""

	def __init__(self, fd):
		self.fd = fd
		self.frame = -1
		self.fd.writelines('# Region file format: DS9 version 4.0\n')

	def select_frame(self, i):
		si = str(i)
		if si != self.frame:
			self.fd.writelines('# frame %s\n' % i)
			self.frame = si


	def add_region(self, region):
		self.fd.writelines( [ str(region), '\n'] )


	def comment(self, s):
		self.fd.writelines( ['# ', s, '\n'] )

	def flush(self):
		return self.fd.flush()

	def plot(self, name, *args, **kwargs):
		self.comment('plot %s' % name)


	def list_plots(self):
		return []

def test_io():
	print "TIO"
	ds9_io()


def test_tags():
	print "TT"
	tmp = ds9()
	tmp.add_region(line(1,1,50,50, tags=('aa', 't2', 'xyz'), text='foo bar') )
	for r in tmp.get_regions():
		if r.getprop('text', '') == 'foo bar':
			assert r.x0 == 1 and r.y0==1 and r.x1==50 and r.y1==50
			assert 'aa' in r.getprop('tags')
			assert 't2' in r.getprop('tags')
			assert 'xyz' in r.getprop('tags')


def test_plot():
	print "TPL"
	import math
	xylist1 = [ (x, math.cos(x*0.3)) for x in range(100) ]
	xylist2 = [ (x, math.sin(x*0.33)) for x in range(100) ]
	tmp = ds9()
	tmp.plot('some_data', xylist1, 'Some Data', 'x-axis', 'y-axis')
	tmp.plot('some_data', xylist2, color='red')
	time.sleep(6)



if __name__ == '__main__':
	import sys
	if len(sys.argv) == 2:
		for r in read_regions_iter(open(sys.argv[1], 'r')):
			print r
		sys.exit(0)
	test_property_parser()
	test_text_kluge()
	test_region_parser()
	test_io()
	test_tags()
	test_plot()




from __future__ import division


class _keyed_thing:
	def __init__(self):
		self._s = {}
	
	def keys(self):
		return self._s.keys()
	
	def values(self):
		return [ self._out(v) for v in self._s.itervalues() ]
	
	def iterkeys(self):
		return self._s.iterkeys()
	
	def items(self):
		return [ (k, self._out(v)) for (k, v) in self._s.iteritems() ]
	
	def get(self, key):
		return self._out( self._s[key] )
	
	def __contains__(self, key):
		return key in self._s
	
	def clear(self):
		return self._s.clear()
	
	def has_key(self):
		return self._s.has_key()
	
	def __iter__(self):
		return self._s.iterkeys()
	
	def __len__(self):
		return len(self._s)
	
	def __delitem__(self, k):
		del self._s[k]
	


class keyed_avg(_keyed_thing):
	def __init__(self):
		_keyed_thing.__init__(self)
	
	def add(self, key, value):
		try:
			s, n = self._s[key]
		except KeyError:
			s = 0.0
			n = 0
		self._s[key] = (s+value, n+1)
	
	@staticmethod
	def _out(stored):
		s, n = stored
		return s/n


def test():
	x = keyed_avg()
	x.add('a', 1)
	x.add('a', 2)
	x.add('b', 2)
	print x.get('a')
	print x.keys()
	print x.values()

if __name__ == '__main__':
	test()



#!/usr/bin/env python

"""Simple centered Gaussian Classifier.
It's designed for a case where there is a lot of data and
where the data is a multivariate Gaussian.
"""


import die
import Num
import q_classifier_r as QCR
import dictops
import fiatio
import gpk_writer

COVERAGE = 3
FTEST = 0.25
Fudge = 1e-6
PSYCO = True

def logdet(cov):
	"""Returns the log of the determinant of a matrix."""
	q = Num.LA.eigvalsh(cov)
	assert Num.alltrue(q > 0)
	return Num.sum(Num.log(q))


def typ_trace(covlist):
	"""This raises all the eigenvalues a little bit,
	so that processing can proceed, even if the covariance
	is singular.   To first order, the fudge gets subtracted
	later on in the processing.
	"""
	tr = 0.0
	dim = None
	for (k, cov) in covlist.items():
		tr += Num.trace(cov)/cov.shape[0]
		if dim is None:
			dim = cov.shape[0]
		else:
			assert cov.shape[0] == dim
	tr /= len(covlist)
	return tr


def boost_diag(covlist, diagboost):
	"""This raises all the eigenvalues a little bit,
	so that processing can proceed, even if the covariance
	is singular.   To first order, the fudge gets subtracted
	later on in the processing.
	"""
	for c in covlist.values():
		trm = Num.identity(c.shape[0])*diagboost
		Num.add(c, trm, c)




def go_auto(fd, Mu=False, coverage=COVERAGE, ftest=FTEST, cb=None, cfd_file = None):
	d = QCR.read_data(fd)
	if cfd_file == None:
		classout = gpk_writer.null_writer()
	else:
		classout = fiatio.writer(open(cfd_file, 'w'))
	summary, out, wrong = QCR.compute_self_class(d, coverage=coverage, ftest=ftest,
						n_per_dim = 0,
						modelchoice = None,
						builder=cb,
						classout=classout)
	QCR.default_writer(summary, out, classout, wrong)



def _do_test(d, coverage=COVERAGE, ftest=FTEST, cb=None):
	classout = gpk_writer.null_writer()
	summary, out, wrong = QCR.compute_self_class(d, coverage=coverage, ftest=ftest,
						n_per_dim = 0,
						modelchoice = None,
						builder=cb,
						classout=classout)
	print 'Summary=', summary
	return summary



def n_zero_cov_builder(d, N, modelchoice=None, trainingset_name=None,
			diagboost=Fudge):
	# ignore N
	assert modelchoice is None
	die.info('Mu_cov')
	assert len(d) > 0, "No data!"
	dim = d[0].value.shape[0]
	covlist = dictops.dict_of_X( lambda v: Num.array(v, Num.Float, copy=True),
					lambda acc, delta: Num.add(acc, delta, acc)
					)
	nlist = dictops.dict_of_accums()
	for x in d:
		nlist.add(x.classid, 1)
		covlist.add(x.classid, Num.outerproduct(x.value, x.value))

	for k in covlist.keys():
		# print 'k=', k
		Num.divide(covlist[k], nlist[k], covlist[k])
		# print 'Eigenval:a:', Num.LA.eigvalsh(covlist[k])

	boost_diag(covlist, diagboost*typ_trace(covlist))

	# We can't subtract the two inverse covariances here,
	# otherwise we mess up the calculation of logdet().

	models = {}
	for k in covlist.keys():
		models[k] = QCR.qmodel(Num.zeros((dim,), Num.Float),
					Num.LA.inv(covlist[k]),
					-logdet(covlist[k])
					)
	

	die.info('Mu_cov done')	
	return [ QCR.classifier('quadratic_class_model', models, trainingset_name=trainingset_name,
				uid=trainingset_name
				)
			]


def _droplist1(dim):
	return [ None ] + [ (i,) for i in range(dim) ]

def _droplist2s(dim):
	return [ None ] + [ (i,(i+1)%dim) for i in range(dim) ]

def _droplist2x(dim):
	return [ None ] + [ (i,j) for i in range(dim) for j in range(dim)]



def n_zero_cov_builder_drop(d, N, modelchoice=None, trainingset_name=None,
				diagboost=Fudge):
	"""Needs to return a list, not an iterator."""
	# ignore N
	assert modelchoice is None
	die.info('Mu_cov')
	assert len(d) > 0, "No data!"
	dim = d[0].value.shape[0]

	def _drop(v, droplist):
		tmp = Num.array(v, copy=True)
		if droplist is not None:
			for i in droplist:
				tmp[i] = 0.0
		return tmp

	typ_tr = None
	o = []
	global Droplist
	for drop in Droplist(dim):
		covlist = dictops.dict_of_X( lambda v: Num.array(v, Num.Float, copy=True),
						lambda acc, delta: Num.add(acc, delta, acc)
						)
		nlist = dictops.dict_of_accums()
		for x in d:
			nlist.add(x.classid, 1)
			xvtmp = _drop(x.value, drop)
			covlist.add(x.classid, Num.outerproduct(xvtmp, xvtmp))
	
		for k in covlist.keys():
			# print 'k=', k
			Num.divide(covlist[k], nlist[k], covlist[k])
			# print 'Eigenval:a:', Num.LA.eigvalsh(covlist[k])
	
		if typ_tr is None:
			# We need to boost the diagonal by the same amount
			# for each different dropped component, otherwise
			# the discriminants of the inverse covariance matrix
			# will be affected in bad ways.
			typ_tr = typ_trace(covlist)
		boost_diag(covlist, diagboost*typ_tr)
	
		# We can't subtract the two inverse covariances here,
		# otherwise we mess up the calculation of logdet().
	
		models = {}
		for k in covlist.keys():
			models[k] = QCR.qmodel(Num.zeros((dim,), Num.Float),
						Num.LA.inv(covlist[k]),
						-logdet(covlist[k])
						)
		
	
		tmp = QCR.classifier('quadratic_class_model', models,
						trainingset_name=trainingset_name,
				uid=hash((trainingset_name, drop))
				)
		tmp.add_info('drop_component', drop)
		o.append( tmp )
	return o



def n_mu_cov_builder(d, N, modelchoice=None, trainingset_name=None,
			diagboost=Fudge):
	# ignore N
	assert modelchoice is None
	die.info('Mu_cov')
	dim = d[0].value.shape[0]
	mulist = dictops.dict_of_X( lambda v: Num.array(v, Num.Float),
					lambda acc, delta: Num.add(acc, delta, acc)
					)
	covlist = dictops.dict_of_X( lambda v: Num.array(v, Num.Float),
					lambda acc, delta: Num.add(acc, delta, acc)
					)
	nlist = dictops.dict_of_accums()
	for x in d:
		nlist.add(x.classid, 1)
	for x in d:
		mulist.add(x.classid, x.value)
	for k in mulist.keys():
		Num.divide(mulist[k], nlist[k], mulist[k])
	mdof = dim
	die.info('Mu done')
	for x in d:
		tmp = x - mulist[x.classid]
		covlist.add(x.classid, Num.outerproduct(tmp, tmp))
	for k in mulist.keys():
		# print 'k=', k
		assert nlist[k] > mdof
		Num.divide(covlist[k], nlist[k]-mdof, covlist[k])
		# print 'Eigenval:a:', Num.LA.eigvalsh(covlist[k])

	for k in mulist.keys():
		# print 'k=', k
		Num.divide(covlist[k], nlist[k], covlist[k])
		# print 'Eigenval:a:', Num.LA.eigvalsh(covlist[k])

	boost_diag(covlist, diagboost)


	models = {}
	for k in mulist.keys():
		models[k] = QCR.qmodel(mulist[k],
					Num.LA.inv(covlist[k]),
					logdet(covlist[k])
					)
	

	die.info('Mu_cov done')	
	return [QCR.classifier('quadratic_class_model', models, trainingset_name=trainingset_name,
				uid=trainingset_name
				)
			]



def _test(dim=3, f=10, low=0.90, high=1.00, N=1000, u=False):
	import math
	if u:
		tmp1 = Num.RA.standard_normal((dim,dim))
		umat1, s, vt = Num.LA.svd( Num.dot(tmp1, Num.transpose(tmp1)) )
		tmp2 = Num.RA.standard_normal((dim,dim))
		umat2, s, vt = Num.LA.svd( Num.dot(tmp2, Num.transpose(tmp2)) )
	d = []
	for i in range(N):
		vector = Num.RA.standard_normal((dim,))
		if u:
			vector = Num.dot(vector, umat1)
		d.append( QCR.datum_tr(vector, 'a') )
	for i in range(N):
		vector = Num.RA.standard_normal((dim,))
		Num.multiply(vector,
				Num.exp((math.log(f)/dim)*(1+Num.arange(dim))),
				vector)
		if u:
			vector = Num.dot(vector, umat2)
		d.append( QCR.datum_tr(vector, 'b') )
	summary = _do_test(d, cb=Cb)
	assert low <= summary['Pcorrect'] <= high


def test1():
	_test(dim=3, f=10, low=0.90, high=1.00)
	_test(dim=3, f=0.1, low=0.90, high=1.00)
	_test(dim=6, f=30, low=0.97, high=1.00)
	_test(dim=3, f=1.0, low=0.30, high=0.70)
	_test(dim=6, f=30, low=0.97, high=1.00, u=True)
	_test(dim=9, f=10, low=0.97, high=1.00, u=True)
	_test(dim=9, f=0.1, low=0.97, high=1.00, u=True)
	_test(dim=2, f=1.0, low=0.30, high=0.70, u=True)
	_test(dim=12, f=1.0, low=0.30, high=0.70, u=True)
	_test(dim=1, f=2.0, low=0.60, high=0.85)
	_test(dim=1, f=1.5, low=0.50, high=0.8)
	_test(dim=1, f=2.5, low=0.60, high=0.85)
	_test(dim=1, f=3.5, low=0.65, high=0.90)
	_test(dim=1, f=5.0, low=0.75, high=0.96)

PSYCO = True

if __name__ == '__main__':
	import sys

	if PSYCO:
		try:
			import psyco
			psyco.full()
		except ImportError:
			pass

	arglist = sys.argv[1:]
	Cb = n_zero_cov_builder
	cfd_file = 'classified.fiat'
	Droplist = _droplist1
	while arglist and arglist[0].startswith('-'):
		arg = arglist.pop(0)
		if arg == '--':
			break
		elif arg == '-coverage':
			COVERAGE = float(arglist.pop(0))
		elif arg == '-drop' or arg == '-drop1':
			Droplist = _droplist1
			Cb = n_zero_cov_builder_drop
		elif arg == '-drop2s':
			Droplist = _droplist2s
			Cb = n_zero_cov_builder_drop
		elif arg == '-drop2x':
			Droplist = _droplist2x
			Cb = n_zero_cov_builder_drop
		elif arg == '-nodetails':
			cfd_file = None
		elif arg == '-test':
			test1()
			sys.exit(0)
		elif arg == '-ftest':
			FTEST = float(arglist.pop(0))
			assert 0.0 < FTEST < 1.0
		else:
			die.die('Unrecognized argument: %s' % arg)
	go_auto(sys.stdin, coverage=COVERAGE, ftest=FTEST, cb=Cb,
			cfd_file = cfd_file)





def beamsearch(cost, extra, initial, B, E):
	"""A breadth-first beam search.
	B = max number of options to keep,
	E = max cost difference between best and worst threads in beam.
	initial = [ starting positions ]
	extra = arbitrary information for cost function.
	cost = fn(state, extra) -> (total_cost, [next states], output_if_goal)
	"""

	o = []
	B = max(B, len(initial))
	hlist = [ (0.0, tmp) for tmp in initial ]
	while len(hlist)>0:
		# print "Len(hlist)=", len(hlist), "len(o)=", len(o)
		hlist.sort()
		if len(hlist) > B:
			hlist = hlist[:B]
		# print "E=", hlist[0][0], " to ", hlist[0][0]+E
		hlist = filter(lambda q, e0=hlist[0][0], e=E: q[0]-e0<=e, hlist)
		# print "		after: Len(hlist)=", len(hlist)
		nlist = []
		while len(hlist) > 0:
			c, point = hlist.pop(0)
			newcost, nextsteps, is_goal = cost(point, extra)
			if is_goal:
				o.append((newcost, is_goal))
			for t in nextsteps:
				nlist.append((newcost, t))
		hlist = nlist
	o.sort()
	return o

	



"""This module has functions that help you dynamically import modules.
"""



import sys
import imp
import os.path
from gmisclib import die

def split_name(name):
	"""Split a name in the form    a/b.c into a, b, c, where
	a is a search path,
	b is a module (package) name, and
	c is a name in the module.
	"""
	p = None
	fcn = []
	if '/' in name:
		p, name = os.path.split(name)
	if '.' in name:
		a = name.split('.')
		name = a[0]
		fcn = a[1:]
	return (p, name, fcn)


def load(name, path):
	"""Load a module from the specified path, or, failing that,
	look in sys.path then for a builtin.   It returns the module,
	but does not import it.
	If path is None, only look in sys.path and builtins.
	If path is an array containing None, replace the None with sys.path.

	This version gives you detailed control over the search path.
	"""
	if path is None:
		pth = None
	else:
		pth = []
		for d in path:
			if d is None:
				pth.extend(sys.path)
			else:
				pth.append( d )
	fd = None
	imp.acquire_lock()
	try:
		fd, pn, desc = imp.find_module(name, pth)
	except ImportError:
		try:
			fd, pn, desc = imp.find_module(name, None)
		except:
			die.warn('#ImportError in load of "%s" from path=%s' % (name, pth))
			imp.release_lock()
			raise

	# If the same module has already been imported, use the existing version.
	if(name in sys.modules and
		hasattr(sys.modules[name], '__file__') and
		os.path.dirname(sys.modules[name].__file__) == os.path.dirname(pn)
			):
		fd.close()
		tmp = sys.modules[name]
		imp.release_lock()
		return tmp

	# print 'name', name, fd, pn, desc
	try:
		pymod = imp.load_module(name, fd, pn, desc)
	finally:
		if fd:
			fd.close()
		imp.release_lock()
	return pymod


load_mod = load
load_inc_path = load
load_mod_inc_path = load_inc_path

def load_named(name, use_sys_path=True):
	"""Load a module.
	If the module name is in the form a/b,
	it looks in directory "a" first.
	If use_sys_path is true, it searches the entire Python path
	
	It returns the module, but does not import it.
	This version handles importing packages and functions nicely,
	but with less control over the search path.
	
	Usage:
		- load_named_module('/dir/my_module'), or
		- load_named_module('foo/my_module'), or
		- load_named_module('foo/my_module.submodule.function'), or
		- various combinations.
	"""
	p, lname, attrlist = split_name(name)
	# print '#', name, '->', lname
	if use_sys_path:
		path = sys.path[:]
	else:
		path = []
	if p is not None:
		path.insert(0, p)
	try:
		tmp0 = load(lname, path)
	except ImportError, x:
		raise ImportError, "%s in directory '%s'%s" % (str(x), p, ['', ' or sys.path'][use_sys_path])

	tmp = tmp0
	try:
		for a in attrlist:
			tmp = getattr(tmp, a)
	except AttributeError:
		filename = getattr(tmp0, '__file__', None)
		if filename is None:
			filename = getattr(tmp0, '__name__', '???')
		die.info("Looking for %s amidst %s" % (a, dir(tmp)))
		raise ImportError, "Lookup fails after import of %s from %s: attempting %s of %s" % (
					lname, filename, a, '.'.join(attrlist)
					)
	return tmp


load_named_module = load_named
load_named_fcn = load_named
load_fcn = load_named

def _test():
	assert abs(load_named('math.pi') - 3.14159) < 0.001
	assert load_named('load_mod.load_named.__doc__') == load_named.__doc__
	f = os.path.splitext(die.__file__)[0]
	assert load_named('%s' % f).__doc__ == die.__doc__

if __name__ == '__main__':
	_test()



import lapack_dsy
import Numeric


def _lg(n):
	k = 0
	twok = 1
	while twok < n:
		k += 1
		twok *= 2
	return k

def dsyevd(a   , jobz = 'V', uplo='U'):
	"""Eigenvalues and eigenvectors of a real symmetric matrix."""
	if jobz != 'V':
		jobz = 'N'
	a = Numeric.array(a, Numeric.Float)
	n = a.shape[0]
	assert a.shape == (n, n)
	eval = Numeric.zeros((n,), Numeric.Float)
	# work = Numeric.zeros((2*(1+5*n+2*n*_lg(n)+3*n*n),), Numeric.Float)
	work = Numeric.zeros((1+5*n+2*n*_lg(n)+3*n*n,), Numeric.Float)
	iwork = Numeric.zeros((2*(2+5*n),), Numeric.Int)
	info = 0
	o = lapack_dsy.dsyevd(jobz, uplo, n, a,
				n, eval, work, len(work),
				iwork, len(iwork), info)
	assert o['info'] == 0
	if jobz != 'V':
		a = None
	return (eval, a)



def _err(a):
	# return Numeric.sum(Numeric.ravel(a)**2)
	return Numeric.sum(a**2)

def _asdiagonal(a):
	n = a.shape[0]
	return Numeric.identity(n)*a

def _reconstruct(eval, evec):
	x = Numeric.matrixmultiply(Numeric.transpose(a2),
				Numeric.matrixmultiply(_asdiagonal(ev2),
					a2))
	return x


if __name__ == '__main__':
	i = Numeric.array([[2, 0], [0, 1]], Numeric.Float)
	print dsyevd(i)
	import RandomArray
	i = RandomArray.random((4,4)) + 1*Numeric.identity(4)
	i = i + Numeric.transpose(i)
	eval, a = dsyevd(i)
	import LinearAlgebra
	ev2, a2 = LinearAlgebra.eigenvectors(i)

	x = _reconstruct(ev2, a2)
	assert _err(x-i) < 1e-8
	y = _reconstruct(eval, a)
	assert _err(y-i) < 1e-8
	assert _err(x-y) < 1e-8



import g_pipe
import Num

LREG = "/home/gpk/misctts/lreg_fit2"

def fill(f, wt, s):
	assert len(f)==len(wt)
	pi, po = g_pipe.popen2(LREG, ["lreg_fit2", "-smooth", "%g" % s])

	for tfw in zip(f, wt):
		pi.write("%g %g\n" % tfw)
	pi.close()
	o = map(float, po.readlines())
	po.close()
	assert len(o) == len(f)
	tmp = Num.asarray(o, Num.Float)
	return tmp

